from openobd_protocol.Messages import Empty_pb2 as grpcEmpty
from openobd_protocol.SessionController.Messages import SessionController_pb2 as grpcSessionController

from openobd.core._token import Token
from openobd.core.exceptions import raises_openobd_exceptions
from openobd.core.grpc_factory import NetworkGrpcFactory
from openobd.core.utils import _get_value_from_kwargs_or_env

class OpenOBDSessionController:

    def __init__(self, **kwargs):
        """
        Used exclusively for starting and managing openOBD sessions. Retrieves the Partner API credentials from the
        environment variables unless explicitly given as kwargs.

        :keyword client_id: the identifier of the created credential set.
        :keyword client_secret: the secret of the created credential set.
        :keyword cluster_id: the ID of the cluster on which openOBD sessions should be managed (001=Europe, 002=USA).
        :keyword grpc_host: the address to which gRPC calls should be sent.
        :keyword grpc_port: the port used by gRPC calls, which needs to be 443 to use SSL.
        """
        self.client_id = _get_value_from_kwargs_or_env(kwargs, "client_id", "OPENOBD_PARTNER_CLIENT_ID")
        self.client_secret = _get_value_from_kwargs_or_env(kwargs, "client_secret", "OPENOBD_PARTNER_CLIENT_SECRET")
        self.cluster_id = _get_value_from_kwargs_or_env(kwargs, "cluster_id", "OPENOBD_CLUSTER_ID")

        grpc_host = _get_value_from_kwargs_or_env(kwargs, "grpc_host", "OPENOBD_GRPC_HOST")
        grpc_port = kwargs.get('grpc_port') if 'grpc_port' in kwargs else 443

        grpc_factory = NetworkGrpcFactory(grpc_host, grpc_port)

        self.session_controller = grpc_factory.get_session_controller()
        self.session_controller_token = Token(self._request_session_controller_token, 7200)

    def _metadata(self):
        metadata = [("authorization", "Bearer {}".format(self.session_controller_token.get_value()))]
        metadata = tuple(metadata)
        return metadata

    @raises_openobd_exceptions
    def _request_session_controller_token(self):
        return self.session_controller.getSessionControllerToken(
            grpcSessionController.Authenticate(
                client_id=self.client_id,
                client_secret=self.client_secret,
                cluster_id=self.cluster_id
            )
        ).value

    @raises_openobd_exceptions
    def start_session_on_ticket(self, ticket_id: grpcSessionController.TicketId) -> grpcSessionController.SessionInfo:
        """
        Starts an openOBD session on the given ticket.

        :param ticket_id: the ticket number (or identifier) on which a session should be started.
        :return: a SessionInfo object representing the started session.
        """
        return self.session_controller.startSessionOnTicket(ticket_id, metadata=self._metadata())

    @raises_openobd_exceptions
    def start_session_on_connector(self, connector_id: grpcSessionController.ConnectorId) -> grpcSessionController.SessionInfo:
        """
        Starts an openOBD session on the given connector.

        :param connector_id: the UUID of the connector on which a session should be started.
        :return: a SessionInfo object representing the started session.
        """
        return self.session_controller.startSessionOnConnector(connector_id, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Retrieves the requested openOBD session. Raises an OpenOBDException if the session does not exist.

        :param session_id: the identifier of the session to be retrieved.
        :return: a SessionInfo object representing the requested session.
        """
        return self.session_controller.getSession(session_id, metadata=self._metadata())

    @raises_openobd_exceptions
    def interrupt_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Forcefully closes the given openOBD session. This changes the session's state to "interrupted" and prevents
        further communication with the session.

        :param session_id: the identifier of the session to be interrupted.
        :return: a SessionInfo object representing the interrupted session.
        """
        return self.session_controller.interruptSession(session_id, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_session_list(self) -> grpcSessionController.SessionInfoList:
        """
        Retrieves all (recently) active openOBD sessions for this partner.

        :return: a SessionInfoList object containing an iterable of SessionInfo objects under its "sessions" attribute.
        """
        return self.session_controller.getSessionList(request=grpcEmpty.EmptyMessage(), metadata=self._metadata())



# Copyright (c) 2024 Jifeline Networks B.V.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""openOBD Python Client implementation"""

try:
    # pylint: disable=ungrouped-imports
    from openobd.core import __version__
except ImportError:
    __version__ = "dev0"


'''Import the message definitions from openobd_protocol '''

'''Import the message definitions from openobd_protocol '''
from openobd_protocol.Communication.Messages.Isotp_pb2 import *
from openobd_protocol.Communication.Messages.Raw_pb2 import *
from openobd_protocol.Communication.Messages.Kline_pb2 import *
from openobd_protocol.Communication.Messages.Terminal15_pb2 import *
from openobd_protocol.Communication.Messages.Doip_pb2 import *
from openobd_protocol.Configuration.Messages.BusConfiguration_pb2 import *
from openobd_protocol.Configuration.Messages.CanBus_pb2 import *
from openobd_protocol.Configuration.Messages.KlineBus_pb2 import *
from openobd_protocol.Configuration.Messages.Terminal15Bus_pb2 import *
from openobd_protocol.Configuration.Messages.DoipBus_pb2 import *
from openobd_protocol.ConnectionMonitor.Messages.ConnectorInformation_pb2 import *
from openobd_protocol.Messages.Empty_pb2 import *
from openobd_protocol.Session.Messages.ServiceResult_pb2 import *
from openobd_protocol.SessionController.Messages.SessionController_pb2 import *
from openobd_protocol.UserInterface.Messages.UserInterface_pb2 import *
from openobd_protocol.Function.Messages.Function_pb2 import *
from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import *
from openobd_protocol.VehicleInfo.Messages.VehicleInfo_pb2 import *

'''Import all classes in subpackages for convenience'''
from .core import *
from .functions import *
from .communication import *
from .ui import *



from enum import Enum
from typing import Iterator

from openobd_protocol.Communication.Messages import Isotp_pb2 as grpcIsotp, Raw_pb2 as grpcRaw, Kline_pb2 as grpcKline, \
    Terminal15_pb2 as grpcTerminal15, Doip_pb2 as grpcDoip
from openobd_protocol.Configuration.Messages import BusConfiguration_pb2 as grpcBusConfiguration
from openobd_protocol.Configuration.Messages import Configuration_pb2 as grpcConfiguration
from openobd_protocol.ConnectionMonitor.Messages import ConnectorInformation_pb2 as grpcConnectionInformation
from openobd_protocol.Function.Messages import Function_pb2 as grpcFunction
from openobd_protocol.Function.Messages.Function_pb2 import Variable
from openobd_protocol.Messages import Empty_pb2 as grpcEmpty
from openobd_protocol.Messages.Empty_pb2 import EmptyMessage
from openobd_protocol.Session.Messages import Session_pb2 as grpcSession, ServiceResult_pb2 as grpcServiceResult
from openobd_protocol.SessionController.Messages import SessionController_pb2 as grpcSessionController
from openobd_protocol.UserInterface.Messages import UserInterface_pb2 as grpcUserInterface
from openobd_protocol.VehicleInfo.Messages import VehicleInfo_pb2 as grpcVehicleInfo

from openobd.core.exceptions import raises_openobd_exceptions
from openobd.core.grpc_factory import GrpcFactory, NetworkGrpcFactory


class ContextVariableType(Enum):
    FUNCTION = 1
    GLOBAL = 2
    CONNECTION = 3

class OpenOBDSession:

    AUTH_TOKEN      = 1
    MONITOR_TOKEN   = 2
    SESSION_TOKEN   = 3

    def __init__(self, session_info: grpcSessionController.SessionInfo, grpc_port=443, grpc_factory: GrpcFactory = None):
        """
        An object that represents an openOBD session. Can be used to make session-specific gRPC calls.

        :param session_info: a SessionInfo object received from a gRPC call.
        :param grpc_port: the default value of 443 should be used to make gRPC calls using SSL.
        :param grpc_factory: GrpcFactory implementation that delegates grpc calls to the desired implementation. Defaults
                             to the NetworkGrpcFactory, which will make grpc calls to the server, specified in the session_info object.
        """
        self.active = True
        self.session_info = session_info

        '''Initially the session context is not initialized (will be initialized when starting a new context)'''
        self.function_context = None    # type: grpcFunction.FunctionContext | None

        '''Initially the session token is the authentication token for the session'''
        self.session_token = self.session_info.authentication_token

        '''Initialize a grpc channel based on the information provided, if no channel is given'''
        if grpc_factory is None:
            grpc_factory = NetworkGrpcFactory(self.session_info.grpc_endpoint, grpc_port)

        self.session = grpc_factory.get_session()
        self.function = grpc_factory.get_function()
        self.config = grpc_factory.get_config()
        self.can = grpc_factory.get_can()
        self.kline = grpc_factory.get_kline()
        self.terminal15 = grpc_factory.get_terminal15()
        self.doip = grpc_factory.get_doip()
        self.ui = grpc_factory.get_ui()
        self.connector_monitor = grpc_factory.get_connection_monitor()
        self.vehicle_info = grpc_factory.get_vehicle_info()

    def id(self):
        return self.session_info.id

    def _metadata(self, token=AUTH_TOKEN):
        bearer_token = ""
        if token == self.AUTH_TOKEN:
            bearer_token = self.session_token
        elif token == self.MONITOR_TOKEN:
            if self.function_context is not None:
                bearer_token = self.function_context.monitor_token
        elif token == self.SESSION_TOKEN:
            bearer_token = self.session_token

        '''Construct the metadata for the gRPC call'''
        metadata = []
        metadata.append(("authorization", "Bearer {}".format(bearer_token)))
        metadata = tuple(metadata)
        return metadata

    def update_session_token(self, session_token):
        """
        Replaces the session token used for gRPC calls.

        :param session_token: the new session token to use.
        """
        self.session_token = session_token

    @raises_openobd_exceptions
    def configure_bus(self, bus_configurations: Iterator[grpcBusConfiguration.BusConfiguration]) -> grpcEmpty.EmptyMessage:
        """
        Configures all given buses so that they can be used for communication. Overwrites any previous bus
        configurations.

        :param bus_configurations: BusConfiguration messages representing the buses that need to be configured.
        :return: an EmptyMessage, indicating that no problems occurred.
        """
        return self.config.configureBus(bus_configurations, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_configuration(self) -> grpcConfiguration.Configuration:
        """
        Returns information on all currently configured buses and channels

        :return: an EmptyMessage, indicating that no problems occurred.
        """
        return self.config.getConfiguration(EmptyMessage(), metadata=self._metadata())

    @raises_openobd_exceptions
    def open_isotp_stream(self, isotp_messages: Iterator[grpcIsotp.IsotpMessage]) -> Iterator[grpcIsotp.IsotpMessage]:
        """
        Opens a bidirectional stream for ISO-TP communication with the channel specified in the given IsotpMessage.

        :param isotp_messages: each IsotpMessage that should be sent to the specified channel.
        :return: IsotpMessages sent by the specified channel.
        """
        return self.can.openIsotpStream(isotp_messages, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_raw_stream(self, raw_frames: Iterator[grpcRaw.RawFrame]) -> Iterator[grpcRaw.RawFrame]:
        """
        Opens a bidirectional stream for raw frame communication with the channel specified in the given RawFrame.

        :param raw_frames: each RawFrame that should be sent to the specified channel.
        :return: RawFrames sent by the specified channel.
        """
        return self.can.openRawStream(raw_frames, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_kline_stream(self, kline_messages: Iterator[grpcKline.KlineMessage]) -> Iterator[grpcKline.KlineMessage]:
        """
        Opens a bidirectional stream for K-Line communication with the channel specified in the given KlineMessage.

        :param kline_messages: each KlineMessage that should be sent to the specified channel.
        :return: KlineMessages sent by the specified channel.
        """
        return self.kline.openKlineStream(kline_messages, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_terminal15_stream(self, empty_message: EmptyMessage) -> Iterator[grpcTerminal15.Terminal15Message]:
        """
        Opens a stream for Terminal15 communication.

        :return: Terminal15 messages sent by the remote.
        """
        return self.terminal15.openTerminal15Stream(empty_message, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_doip_stream(self, doip_messages: Iterator[grpcDoip.DoipMessage]) -> Iterator[grpcDoip.DoipMessage]:
        """
        Opens a bidirectional stream for DoIP communication with the channel specified in the given DoipMessage.

        :param doip_messages: each DoipMessage that should be sent to the specified channel.
        :return: DoipMessages sent by the specified channel.
        """
        return self.doip.openDoipStream(doip_messages, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_control_stream(self, user_interface_messages: Iterator[grpcUserInterface.Control]) -> Iterator[grpcUserInterface.Control]:
        """
        Opens a stream that displays given Control messages to the customer or operator, and returns their response.

        :param user_interface_messages: Control messages that need to be displayed on the user interface.
        :return: Control messages containing the user's response, depending on which Control type was sent.
        """
        return self.ui.openControlStream(user_interface_messages, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_connector_information(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcConnectionInformation.ConnectorInformation:
        """
        Retrieves information on the current status of the connection with the connector.

        :return: a ConnectorInformation message containing the current status of the connection.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.connector_monitor.getConnectorInformation(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_connector_information_stream(self, request: grpcEmpty.EmptyMessage | None = None) -> Iterator[grpcConnectionInformation.ConnectorInformation]:
        """
        Opens a stream which receives a ConnectorInformation message each second, containing the status of the
        connection with the connector.

        :return: ConnectorInformation messages containing the current status of the connection.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.connector_monitor.openConnectorInformationStream(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_ecu_list(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcVehicleInfo.EcuList:
        """
        Returns the ECUs that have been detected in the vehicle.

        :return: an EcuList object containing info on each detected ECU.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.vehicle_info.getEcuList(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def authenticate(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcSession.SessionToken:
        """
        Authenticates a newly created session. This needs to be done once for each session and is required to make any
        other gRPC calls for this session.

        :return: a session token, valid for 5 minutes, which is required to make gRPC calls for this session.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.session.authenticate(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def open_session_token_stream(self, request: grpcEmpty.EmptyMessage | None = None) -> Iterator[grpcSession.SessionToken]:
        """
        Starts a stream which receives a new session token every 2 minutes, each of which is valid for 5 minutes. A
        valid session token is required to make gRPC calls.

        :return: session tokens required to keep the session valid.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.session.openSessionTokenStream(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def start_context(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcFunction.FunctionContext:
        """
        Starts a new function context within the session. This creates a new context with its own isolated variables.

        :return: a session context, including a context uuid and context token, valid for 5 minutes
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        function_context = self.function.startFunctionContext(request=request, metadata=self._metadata()) # type: grpcFunction.FunctionContext

        '''Update the session information, we are now running within a context'''
        self.function_context = function_context

        '''
        Update the authentication token for this session, so we could continue with this session when it is desired.
        But in most cases we want to pass on the updated session info to another openOBD function executor, while this
        proces is monitoring (using monitor tokens) when this execution is finished.
        '''
        self.session_token = function_context.authentication_token

        '''Update the session info, so it can be passed on to another openOBD executor'''
        self.session_info.authentication_token = function_context.authentication_token

        return function_context

    @raises_openobd_exceptions
    def set_context_variable(self, key: str, value, variable_type=ContextVariableType.FUNCTION) -> None:
        if variable_type == ContextVariableType.GLOBAL:
            mem_type = grpcFunction.ContextType.GLOBAL_CONTEXT
        elif variable_type == ContextVariableType.CONNECTION:
            mem_type = grpcFunction.ContextType.CONNECTION_CONTEXT
        else:
            mem_type = grpcFunction.ContextType.FUNCTION_CONTEXT

        if isinstance(value, str):
            self.function.setVariable(
                grpcFunction.Variable(type=mem_type, key=key, value=value),
                metadata=self._metadata())
        else:
            self.function.setVariable(
                grpcFunction.Variable(type=mem_type, key=key, object=value),
                metadata=self._metadata())

    @raises_openobd_exceptions
    def get_context_variable(self, key: str, variable_type=ContextVariableType.FUNCTION) -> Variable:
        if variable_type == ContextVariableType.GLOBAL:
            mem_type = grpcFunction.ContextType.GLOBAL_CONTEXT
        elif variable_type == ContextVariableType.CONNECTION:
            mem_type = grpcFunction.ContextType.CONNECTION_CONTEXT
        else:
            mem_type = grpcFunction.ContextType.FUNCTION_CONTEXT

        return self.function.getVariable(
            grpcFunction.Variable(type=mem_type, key=key),
            metadata=self._metadata())

    @raises_openobd_exceptions
    def delete_context_variable(self, key: str, variable_type=ContextVariableType.FUNCTION) -> None:
        if variable_type == ContextVariableType.GLOBAL:
            mem_type = grpcFunction.ContextType.GLOBAL_CONTEXT
        elif variable_type == ContextVariableType.CONNECTION:
            mem_type = grpcFunction.ContextType.CONNECTION_CONTEXT
        else:
            mem_type = grpcFunction.ContextType.FUNCTION_CONTEXT

        self.function.deleteVariable(
            grpcFunction.Variable(type=mem_type, key=key),
            metadata=self._metadata())

    @raises_openobd_exceptions
    def set_function_argument(self, key: str, value):
        self.function.setFunctionArgument(
            grpcFunction.Variable(key=key, value=value),
            metadata=self._metadata())

    @raises_openobd_exceptions
    def set_function_result(self, key: str, value):
        self.function.setFunctionResult(
            grpcFunction.Variable(key=key, value=value),
            metadata=self._metadata())

    @raises_openobd_exceptions
    def monitor_function_context(self, request: grpcFunction.FunctionContext | None = None) -> Iterator[grpcFunction.FunctionContext]:
        """
        Starts a stream which receives a new monitor token every 2 minutes, each of which is valid for 5 minutes. A
        valid monitor token is required to keep monitoring the context until it finishes.

        :return: session context containing monitor tokens and eventually a new authentication token
        """
        if request is None:
            request = grpcFunction.FunctionContext()
        return self.function.monitorFunctionContext(request=request, metadata=self._metadata(token=OpenOBDSession.MONITOR_TOKEN))

    @raises_openobd_exceptions
    def register_function_details(self, function_details: grpcFunction.FunctionDetails) -> grpcEmpty.EmptyMessage:
        """
        Registers information about what function is running in the current context.

        :param function_details: a FunctionDetails message detailing what function is running in this context.
        :return: an EmptyMessage, indicating that no problems occurred.
        """
        return self.function.registerFunctionDetails(request=function_details, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_function_details(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcFunction.FunctionDetails:
        """
        Retrieves the function details that have been registered in the current context.

        :return: the current context's FunctionDetails.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.function.getFunctionDetails(request=request, metadata=self._metadata())

    @raises_openobd_exceptions
    def finish(self, service_result: grpcServiceResult.ServiceResult) -> grpcEmpty.EmptyMessage:
        """
        Gracefully closes the openOBD session by changing the session's state to "finished" and preventing further
        communication with the session. The given ServiceResult indicates the success or failure reason of the executed
        service.

        :param service_result: a ServiceResult message representing the result of the executed service.
        :return: an EmptyMessage, indicating that no problems occurred.
        """
        self.active = False
        return self.session.finish(service_result, metadata=self._metadata())

    def __str__(self):
        return (f"ID: {self.session_info.id}, "
                f"state: {self.session_info.state}, "
                f"created at: {self.session_info.created_at}, "
                f"gRPC endpoint: {self.session_info.grpc_endpoint}, "
                f"authentication token: {self.session_info.authentication_token}")



import time

from openobd_protocol.UserInterface.Messages.UserInterface_pb2 import Label, Options, Continue, YesNo, FreeText, \
    Numbers, InterfaceType, Control, LabelTranslations, Image

from openobd.core import StreamHandler, OpenOBDSession

_control_types = Label | Options | Continue | YesNo | FreeText | Numbers
control_mapping = {
    Label: "control_label",
    Options: "control_options",
    Continue: "control_continue",
    YesNo: "control_yesno",
    FreeText: "control_freetext",
    Numbers: "control_number"
}

class UiHandler:

    def __init__(self, openobd_session: OpenOBDSession, target: InterfaceType = InterfaceType.INTERFACE_USER):
        """
        Handles sending user interface (UI) elements to a gRPC stream and parsing their responses.

        :param openobd_session: an active OpenOBDSession with which to start the UI stream.
        :param target: determines who is able to view the user interface.
        """
        self.stream_handler = StreamHandler(openobd_session.open_control_stream, outgoing_stream=True)
        self._target = target

    def show_ui(self, control: _control_types, translations: LabelTranslations | list[LabelTranslations] = None, image: Image = None) -> None | int | bool | str:
        """
        Displays a control object and waits for a response.

        :param control: the control type to be displayed.
        :param translations: Translations for all labels used in this control.
        :param image: Optional image to show next to the control.
        :return: the user's response, depending on which control type has been displayed.
        """
        assert type(control) in control_mapping, f"Received unsupported control type {type(control)}."

        # Allow for providing a single LabelTranslation object for convenience
        if isinstance(translations, LabelTranslations):
            translations = [translations]

        control_arg = {
            control_mapping[type(control)]: control
        }

        self.stream_handler.send(Control(
            target=self._target,
            translations=translations,
            image=image,
            **control_arg
        ), flush_incoming_messages=True)
        while True:
            response = self.stream_handler.receive()

            # Confirm that the received control type matches the control type sent earlier. If not, ignore it
            control_incoming = self._get_control_from_message(response)
            if isinstance(control_incoming, type(control)):
                if hasattr(control_incoming, "answer"):
                    return control_incoming.answer
                else:
                    return None

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream and the UI, if they are not already closed. A new UiHandler object will have to be
        created to open the UI again.
        """
        self.stream_handler.stop_stream()
        time.sleep(1.5)     # Give the UI time to close before continuing

    @staticmethod
    def _get_control_from_message(message: Control) -> _control_types:
        """
        Retrieves the control type from a given Control message.

        :param message: Control message containing a control type.
        :return: the control type present in the given message.
        """
        for control_type in control_mapping.values():
            if message.HasField(control_type):
                return getattr(message, control_type)



from openobd_protocol.Communication.Messages import Terminal15_pb2 as grpcTerminal15

from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler

class Terminal15Socket:

    def __init__(self, openobd_session: OpenOBDSession, timeout: float | None = 10):
        """
        Handles receiving terminal15 signals from the given channel.

        :param openobd_session: an active OpenOBDSession with which to start the stream.
        :param timeout: time in seconds to wait for an incoming frame before raising a OpenOBDStreamTimeoutException. None will wait forever.
        """
        self.stream_handler = StreamHandler(openobd_session.open_terminal15_stream, outgoing_stream=False)
        self.timeout = timeout

    def receive(self, block: bool = True, timeout: float | None = None) -> str | None:
        """
        Retrieves the oldest pending incoming message. If there are no pending messages, wait until a new message arrives.

        :param block: whether to wait for a new message if there are no pending messages. If False, returns None when no messages are pending.
        :param timeout: time in seconds to wait for a new frame before raising a OpenOBDStreamTimeoutException. Will use self.timeout if None.
        :return: the payload received from the channel.
        """
        timeout = timeout if timeout is not None else self.timeout
        response = self.stream_handler.receive(block, timeout)
        if response is not None:
            if response.state == grpcTerminal15.STATE_ON:
                return "ON"
            elif response.state == grpcTerminal15.STATE_OFF:
                return "OFF"
            else:
                return "UNDEFINED"

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new RawSocket object will have to be created to start
        another stream.
        """
        self.stream_handler.stop_stream()



from openobd_protocol.Communication.Messages.Kline_pb2 import KlineChannel, KlineMessage

from openobd.communication.response_exceptions import *
from openobd.core.session import OpenOBDSession
from openobd.communication.uds_request_handler import UdsRequestHandler
from openobd.core.stream_handler import StreamHandler


class KlineSocket:

    def __init__(self, openobd_session: OpenOBDSession, kline_channel: KlineChannel, timeout: float | None = 10):
        """
        Handles sending ISO 14230 K-Line messages to the given channel, and receiving their responses.

        :param openobd_session: an active OpenOBDSession with which to start the K-Line stream.
        :param kline_channel: the channel which to send messages to.
        :param timeout: time in seconds to wait for each response before raising a NoResponseException. None will wait forever.
        """
        self.stream_handler = StreamHandler(openobd_session.open_kline_stream, outgoing_stream=True)
        self._channel = kline_channel
        self.uds_request_handler = UdsRequestHandler(self.stream_handler, timeout)

        # Pre-emptively send a message with an empty payload to ensure the K-Line channel is set up by the server
        self.stream_handler.send(KlineMessage(channel=self._channel, payload=""))

    def request(self, payload: str, timeout: float | None = None, silent: bool = False, tries: int = 1) -> str | None:
        """
        Sends the given payload and waits for the response. Raises the appropriate ResponseException when receiving a
        negative response.

        :param payload: the bytes to send to the channel.
        :param timeout: time in seconds to wait for a response before raising a NoResponseException. Will use self.timeout if None.
        :param silent: if True, will not raise any ResponseExceptions. Instead, returns the payload of negative responses or None for NoResponseExceptions.
        :param tries: how many times to send the payload until a positive response is received.
        :return: the payload of the received response.
        """
        try:
            message = KlineMessage(channel=self._channel, payload=payload)
            return self.uds_request_handler.request(message, timeout, silent, tries)
        except ResponseException as e:
            # Add channel info to the exception before raising it
            e.request_id = self._channel.ecu_id
            raise e

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new Socket object will have to be created to start another
        stream.
        """
        self.uds_request_handler.stop_stream()



import threading
import time

from openobd_protocol.Session.Messages.ServiceResult_pb2 import ServiceResult
from openobd_protocol.SessionController.Messages.SessionController_pb2 import SessionInfo

from openobd.core.openobd import OpenOBD
from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler, OpenOBDStreamStoppedException, StreamState
from openobd.functions.function_exceptions import *

class FunctionContextHandler:

    def __init__(self, openobd_session: OpenOBDSession):
        """
        Starts a function context in the given session and starts a thread that monitors the context.

        :param openobd_session: the OpenOBDSession on which to start a context.
        """
        self.openobd_session = openobd_session

        '''Start session context'''
        self.openobd_session.start_context()

        '''Keep track of exactly this context that we want to monitor'''
        self.function_context = self.openobd_session.function_context

        '''
        Create stream handler for receiving session context updates
        The argument is the context that we want to monitor, after that we will just listen
        to an incoming stream for updates (until the context is finished)'''
        self.stream_handler = StreamHandler(self.openobd_session.monitor_function_context, request=self.openobd_session.function_context)

        self.refresh_session_context_thread = threading.Thread(target=self._monitor_session_context, daemon=True)
        self.refresh_session_context_thread.start()

    def run_function(self, function_id: str) -> FunctionContext:
        """
        Uses the OpenOBD singleton to run a function in this context and waits until it has finished.
        Raises an OpenOBDFunctionException if the function did not finish successfully.
        The context will always be closed once this method returns.

        :param function_id: the UUID of the function that needs to be executed.
        :return: a FunctionContext object containing info such as the result.
        """
        try:
            OpenOBD().run_function(function_id, self.get_session_info())
        except OpenOBDException as e:
            # Ensure this context is closed before raising an exception
            self._finish_context([Result.RESULT_FAILURE])
            if e.status == 4:  # deadline exceeded
                raise OpenOBDFunctionDeadlineExceededException(function_id=function_id) from e
            elif e.status == 7:  # permission denied
                raise OpenOBDFunctionPermissionDeniedException(function_id=function_id) from e
            elif e.status == 14:  # unavailable
                raise OpenOBDFunctionUnavailableException(function_id=function_id) from e
            else:
                raise e

        function_context = self.wait()
        self._validate_context_result(function_context, function_id)
        return function_context

    @staticmethod
    def _validate_context_result(function_context: FunctionContext, function_id: str = None):
        failure = False
        for result in function_context.service_result.result:
            if result != Result.RESULT_SUCCESS:
                failure = True
                break
        if failure or len(function_context.service_result.result) == 0:
            raise OpenOBDFunctionUnsuccessfulException(function_id=function_id, function_context=function_context)

    def _finish_context(self, results: list[Result]):
        if self.is_active():
            self.openobd_session.finish(ServiceResult(result=results))
            self.wait()     # Wait until the context has fully closed

    def get_session_info(self) -> SessionInfo:
        return self.openobd_session.session_info

    def _monitor_session_context(self):
        try:
            while True:
                function_context = self.stream_handler.receive()    # type: FunctionContext
                self.openobd_session.function_context = function_context
                if function_context.finished:
                    '''Context has finished'''
                    '''Allow authentication again from openOBD session'''
                    self.openobd_session.session_token = function_context.authentication_token
                    break

        except OpenOBDStreamStoppedException:
            pass
        except OpenOBDException as e:
            print(f" [!] openOBD session [{self.openobd_session.id()}] => " +
                  f"stopped monitoring context [{self.openobd_session.function_context.id}] due to an exception.")
            print(e)
        finally:
            self.stream_handler.stop_stream()

    def is_active(self) -> bool:
        return self.stream_handler.stream_state == StreamState.ACTIVE

    def wait(self) -> FunctionContext:
        """
        A blocking method that waits until this context has finished.

        :return: the final FunctionContext object received.
        """
        try:
            while self.is_active():
                time.sleep(1)
        except Exception:
            pass
        finally:
            return self.openobd_session.function_context



from openobd_protocol.ConnectionMonitor.Messages.ConnectorInformation_pb2 import ConnectorInformation

from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler

class ConnectionMonitor:

    def __init__(self, openobd_session: OpenOBDSession, sample_size: int = 10):
        """
        Starts a stream which receives a ConnectorInformation message each second. Saves the most recently received
        ConnectorInformation messages, and allows them to be returned when requested.

        :param openobd_session: an active OpenOBDSession from which to receive connection information.
        :param sample_size: the maximum amount of ConnectorInformation messages that should be saved at a time.
        """
        self.stream_handler = StreamHandler(openobd_session.open_connector_information_stream, outgoing_stream=False)
        self.sample_size = sample_size
        self._connector_info_samples = []

    def get_connector_info_list(self) -> list[ConnectorInformation]:
        """
        Returns the latest ConnectorInformation messages. The amount of returned messages is dependent on this object's
        sample_size.

        :return: a list containing the latest ConnectorInformation messages, ordered from most recent to oldest.
        """
        # Add any new ConnectorInformation messages to the self._connector_info_samples list
        connector_sample = self.stream_handler.receive(block=False)
        while connector_sample is not None:
            self._connector_info_samples.append(connector_sample)
            connector_sample = self.stream_handler.receive(block=False)

        # Make sure the list does not exceed self.sample_size, by taking only the most recent messages
        self._connector_info_samples = self._connector_info_samples[-self.sample_size:]

        # Reverse the list, so the samples are ordered latest to oldest, and return it
        return self._connector_info_samples[::-1]

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new ConnectionMonitor object will have to be created to
        start another stream.
        """
        self.stream_handler.stop_stream()

    @staticmethod
    def print_connector_information_list(connector_information_list: list[ConnectorInformation]) -> None:
        """
        For each given ConnectorInformation message, print the info it contains.

        :param connector_information_list: a list containing the ConnectorInformation messages to be printed.
        """
        for message in connector_information_list:
            print(f" connected:         {message.connected}")
            print(f" connected since:   {message.connected_since}")
            print(f" latency:           {message.latency} ms")
            print(f" healthy:           {message.healthy}\n")



import os

def _get_value_from_kwargs_or_env(kwargs, kwarg_key, env_key):
    if kwarg_key in kwargs:
        return kwargs[kwarg_key]
        return kwargs[kwarg_key]
    elif env_key in os.environ:
        return os.environ[env_key]
    else:
        raise AssertionError(f"Argument \"{kwarg_key}\" could not be found. Pass it explicitly, or ensure it is available as an environment variable named \"{env_key}\".")



import time

from google.protobuf.message import Message

from openobd.communication.response_exceptions import *
from openobd.core.exceptions import OpenOBDStreamTimeoutException
from openobd.core.stream_handler import StreamHandler


class UdsRequestHandler:

    def __init__(self, stream_handler: StreamHandler, timeout: float | None):
        """
        A class that takes care of parsing and handling incoming UDS messages.

        :param stream_handler: a StreamHandler object that manages the stream over which to send messages.
        :param timeout: time in seconds to wait for each response before raising a NoResponseException. None will wait forever.
        """
        self.stream_handler = stream_handler
        self.timeout = timeout

    def request(self, message: Message, timeout: float | None = None, silent: bool = False, tries: int = 1) -> str | None:
        """
        Sends the given message and waits for the response. Raises the appropriate ResponseException when receiving a
        negative response.

        :param message: the gRPC message to send to the channel.
        :param timeout: time in seconds to wait for a response before raising a NoResponseException. Will use self.timeout if None.
        :param silent: if True, will not raise any ResponseExceptions. Instead, returns the payload of negative responses or None for NoResponseExceptions.
        :param tries: how many times to send the message until a positive response is received.
        :return: the payload of the received response.
        """
        assert tries > 0, "Optional argument 'tries' must be greater than zero."

        if hasattr(message, "payload"):
            payload = message.payload
        else:
            raise AttributeError("The given gRPC message must have a 'payload' attribute.")
        timeout = timeout if timeout is not None else self.timeout

        for current_try in range(1, tries + 1):
            self.stream_handler.send(message, flush_incoming_messages=True)

            try:
                end_time = time.time() + timeout if timeout is not None else None
                try:
                    return self._receive_and_handle_response(payload, timeout)
                except BusyRepeatRequestException as e:
                    # We should retry the request after waiting a little
                    if end_time is not None and time.time() + 0.5 > end_time:
                        # We don't have enough time left this try, so raise the received exception
                        raise e
                    time.sleep(0.5)
                    try:
                        # Make another request in the time we have left
                        remaining_time = max(0.0, end_time - time.time()) if end_time is not None else None
                        return self.request(message, timeout=remaining_time)
                    except NoResponseException:
                        # We did not receive another response in time, so raise the exception received earlier
                        raise e

            except ResponseException as e:
                if current_try < tries:
                    continue
                if silent:
                    return e.response
                raise e

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new object will have to be created to start another
        stream.
        """
        self.stream_handler.stop_stream()

    def _receive_and_handle_response(self, request_payload: str, timeout: float | None) -> str:
        end_time = time.time() + timeout if timeout is not None else None

        try:
            response = self.stream_handler.receive(timeout=timeout)
        except OpenOBDStreamTimeoutException:
            no_response_exc = NoResponseException(request=request_payload, response=None)
            raise no_response_exc

        response_payload = response.payload.upper()

        try:
            self._check_response_format(response_payload)

            if not self._response_matches_request(response_payload, request_payload):
                # The received message is not a response to our request, so continue listening in the time we have left
                remaining_time = max(0.0, end_time - time.time()) if end_time is not None else None
                return self._receive_and_handle_response(request_payload, remaining_time)

            try:
                self._check_response_payload(response_payload)
            except RequestCorrectlyReceivedResponsePendingException as e:
                try:
                    # We need to wait a little longer for the response, so continue listening in the time we have left
                    remaining_time = max(0.0, end_time - time.time()) if end_time is not None else None
                    return self._receive_and_handle_response(request_payload, remaining_time)
                except NoResponseException:
                    # Did not receive a timely response, so raise the original ResponsePending exception
                    raise e

        except ResponseException as e:
            # Add the request and response metadata to the exception before raising it
            e.request = request_payload
            # Only set the response attribute if it doesn't already contain a value
            if not e.response:
                e.response = response_payload
            raise e

        return response_payload

    @staticmethod
    def _check_response_format(response: str) -> None:
        """
        Checks whether the response is formatted correctly. If not, raises an InvalidResponseException.

        :param response: the payload of the response.
        """
        if response[:2] == "7F" and len(response) != 6:
            # The response does not have the correct length for a negative response
            raise InvalidResponseException()

    @staticmethod
    def _response_matches_request(response: str, request: str) -> bool:
        """
        Checks whether the service of the response matches the service of the request.

        :param response: the payload of the response.
        :param request: the payload of the request.
        :return: True if the response matches the request. False if there's a mismatch.
        """
        if response == "":
            return False

        try:
            service_byte = int(request[:2], 16)
            if response[:2] == "7F":
                if int(response[2:4], 16) != service_byte:
                    return False
            elif int(response[:2], 16) != (service_byte + 0x40):
                return False
            return True

        except ValueError:
            raise InvalidResponseException()

    @staticmethod
    def _check_response_payload(payload: str) -> None:
        """
        Checks whether the given response is a negative response. If so, raises the appropriate exception.

        :param payload: the payload of the response.
        """
        if payload[:2] != "7F":
            return
        negative_response_code = payload[4:6]
        exception = negative_response_code_exceptions.get(negative_response_code, UnknownNegativeResponseException)
        raise exception



from collections.abc import Iterator

from openobd_protocol.Function.Messages import Function_pb2 as grpcFunction
from openobd_protocol.FunctionBroker.Messages import FunctionBroker_pb2 as grpcFunctionBroker
from openobd_protocol.SessionController.Messages import SessionController_pb2 as grpcSessionController

from openobd.core.exceptions import raises_openobd_exceptions
from openobd.core.session_controller import OpenOBDSessionController
from openobd.core.session import OpenOBDSession
from openobd.core.function_broker import OpenOBDFunctionBroker


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class OpenOBD(metaclass=SingletonMeta):

    def __init__(self, **kwargs):
        """
        A singleton that allows for the starting and managing of openOBD sessions using the provided Partner API
        credentials. The credentials are retrieved from the environment variables unless explicitly given as kwargs
        on first initialization.

        :keyword client_id: the identifier of the created credential set.
        :keyword client_secret: the secret of the created credential set.
        :keyword cluster_id: the ID of the cluster on which openOBD sessions should be managed (001=Europe, 002=USA).
        :keyword grpc_host: the address to which gRPC calls should be sent.
        :keyword grpc_port: the port used by gRPC calls, which needs to be 443 to use SSL.
        """
        self.session_controller = OpenOBDSessionController(**kwargs)
        self.function_broker = OpenOBDFunctionBroker(**kwargs)

    @raises_openobd_exceptions
    def start_session_on_ticket(self, ticket_id: str) -> OpenOBDSession:
        """
        Starts an openOBD session on the given ticket.

        :param ticket_id: the ticket number (or identifier) on which a session should be started.
        :return: an OpenOBDSession object representing the started session, which can be used to make gRPC calls.
        """
        response = self.session_controller.start_session_on_ticket(grpcSessionController.TicketId(value=ticket_id))
        return OpenOBDSession(response)

    @raises_openobd_exceptions
    def start_session_on_connector(self, connector_id: str) -> OpenOBDSession:
        """
        Starts an openOBD session on the given connector.

        :param connector_id: the UUID of the connector on which a session should be started.
        :return: an OpenOBDSession object representing the started session, which can be used to make gRPC calls.
        """
        response = self.session_controller.start_session_on_connector(grpcSessionController.ConnectorId(value=connector_id))
        return OpenOBDSession(response)

    @raises_openobd_exceptions
    def get_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Retrieves the requested openOBD session. Raises an OpenOBDException if the session does not exist.

        :param session_id: the identifier of the session to be retrieved.
        :return: a SessionInfo object representing the requested session.
        """
        return self.session_controller.get_session(session_id)

    @raises_openobd_exceptions
    def interrupt_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Forcefully closes the given openOBD session. This changes the session's state to "interrupted" and prevents
        further communication with the session.

        :param session_id: the identifier of the session to be interrupted.
        :return: a SessionInfo object representing the interrupted session.
        """
        return self.session_controller.interrupt_session(session_id)

    @raises_openobd_exceptions
    def get_session_list(self) -> grpcSessionController.SessionInfoList:
        """
        Retrieves all (recently) active openOBD sessions for this partner.

        :return: a SessionInfoList object containing an iterable of SessionInfo objects under its "sessions" attribute.
        """
        return self.session_controller.get_session_list()

    @raises_openobd_exceptions
    def open_function_stream(self, function_update_messages: Iterator[grpcFunctionBroker.FunctionUpdate]) -> Iterator[grpcFunctionBroker.FunctionUpdate]:
        """
        Opens a stream in which functions can be made available for function callers, and through which function calls
        will be forwarded.

        :param function_update_messages: a FunctionUpdate message for each function that needs to be registered.
        :return: FunctionUpdate messages containing acknowledgements of registrations, pings, and function calls.
        """
        return self.function_broker.open_function_stream(function_update_messages)

    @raises_openobd_exceptions
    def run_function(self, function_id: str, session_info: grpcSessionController.SessionInfo) -> grpcFunctionBroker.FunctionUpdate:
        """
        Execute a function that has been registered by a function launcher.

        :param function_id: the UUID of the function to call.
        :param session_info: the session to run the function in.
        :return: a FunctionUpdate object defining whether the function has been successfully launched.
        """
        return self.function_broker.run_function(grpcFunctionBroker.FunctionCall(id=function_id, session_info=session_info))

    @raises_openobd_exceptions
    def get_function_registration(self, function_id: str) -> grpcFunctionBroker.FunctionRegistration:
        """
        Retrieves information about the requested function. For instance, whether the function is online or not.

        :param function_id: the UUID of the function to request info on.
        :return: a FunctionRegistration object containing details on the requested function.
        """
        return self.function_broker.get_function_registration(grpcFunction.FunctionId(value=function_id))

    @raises_openobd_exceptions
    def generate_function_signature(self) -> grpcFunctionBroker.FunctionSignature:
        """
        Generates a new function ID and signature, which are used when registering an openOBD function.

        :return: a FunctionSignature object containing a new function ID and its corresponding signature.
        """
        return self.function_broker.generate_function_signature()





from typing import Iterator

from openobd_protocol.Function.Messages import Function_pb2 as grpcFunction
from openobd_protocol.FunctionBroker.Messages import FunctionBroker_pb2 as grpcFunctionBroker
from openobd_protocol.Messages import Empty_pb2 as grpcEmpty
from openobd_protocol.SessionController.Messages import SessionController_pb2 as grpcSessionController

from openobd.core._token import Token
from openobd.core.utils import _get_value_from_kwargs_or_env
from openobd.core.grpc_factory import NetworkGrpcFactory
from openobd.core.exceptions import raises_openobd_exceptions

class OpenOBDFunctionBroker:

    def __init__(self, **kwargs):
        """
        Used for hosting and running third-party openOBD functions.
        Retrieves the Partner API credentials from the environment variables unless explicitly given as kwargs.

        :keyword client_id: the identifier of the created credential set.
        :keyword client_secret: the secret of the created credential set.
        :keyword cluster_id: the ID of the cluster on which openOBD functions should be managed (001=Europe, 002=USA).
        :keyword grpc_host: the address to which gRPC calls should be sent.
        :keyword grpc_port: the port used by gRPC calls, which needs to be 443 to use SSL.
        """
        self.client_id = _get_value_from_kwargs_or_env(kwargs, "client_id", "OPENOBD_PARTNER_CLIENT_ID")
        self.client_secret = _get_value_from_kwargs_or_env(kwargs, "client_secret", "OPENOBD_PARTNER_CLIENT_SECRET")
        self.cluster_id = _get_value_from_kwargs_or_env(kwargs, "cluster_id", "OPENOBD_CLUSTER_ID")

        grpc_host = _get_value_from_kwargs_or_env(kwargs, "grpc_host", "OPENOBD_GRPC_HOST")
        grpc_port = kwargs.get('grpc_port') if 'grpc_port' in kwargs else 443

        grpc_factory = NetworkGrpcFactory(grpc_host, grpc_port)

        self.function_broker = grpc_factory.get_function_broker()
        self.function_broker_token = Token(self._request_function_broker_token, 300)

    def _metadata(self):
        metadata = [("authorization", "Bearer {}".format(self.function_broker_token.get_value()))]
        metadata = tuple(metadata)
        return metadata

    @raises_openobd_exceptions
    def _request_function_broker_token(self):
        """
        Requests a new function broker token. A valid function broker token is required to make any of the other calls
        to the function broker.
        """
        return self.function_broker.getFunctionBrokerToken(
            grpcSessionController.Authenticate(
                client_id=self.client_id,
                client_secret=self.client_secret,
                cluster_id=self.cluster_id
            )
        ).value

    @raises_openobd_exceptions
    def open_function_stream(self, function_update_messages: Iterator[grpcFunctionBroker.FunctionUpdate]) -> Iterator[grpcFunctionBroker.FunctionUpdate]:
        """
        Opens a stream in which functions can be made available for function callers, and through which function calls
        will be forwarded.

        :param function_update_messages: a FunctionUpdate message for each function that needs to be registered.
        :return: FunctionUpdate messages containing acknowledgements of registrations, pings, and function calls.
        """
        return self.function_broker.openFunctionStream(function_update_messages, metadata=self._metadata())

    @raises_openobd_exceptions
    def run_function(self, function_call: grpcFunctionBroker.FunctionCall) -> grpcFunctionBroker.FunctionUpdate:
        """
        Execute a function that has been registered by a function launcher.

        :param function_call: a FunctionCall defining which function to call, and the session to run the function in.
        :return: a FunctionUpdate object defining whether the function has been successfully launched.
        """
        return self.function_broker.runFunction(request=function_call, metadata=self._metadata())

    @raises_openobd_exceptions
    def get_function_registration(self, function_id: grpcFunction.FunctionId) -> grpcFunctionBroker.FunctionRegistration:
        """
        Retrieves information about the requested function. For instance, whether the function is online or not.

        :param function_id: the UUID of the function to request info on.
        :return: a FunctionRegistration object containing details on the requested function.
        """
        return self.function_broker.getFunctionRegistration(request=function_id, metadata=self._metadata())

    @raises_openobd_exceptions
    def generate_function_signature(self, request: grpcEmpty.EmptyMessage | None = None) -> grpcFunctionBroker.FunctionSignature:
        """
        Generates a new function ID and signature, which are used when registering an openOBD function.

        :return: a FunctionId object containing a new function ID and its corresponding signature.
        """
        if request is None:
            request = grpcEmpty.EmptyMessage()
        return self.function_broker.generateFunctionSignature(request=request, metadata=self._metadata())



import functools

import grpc


class OpenOBDException(Exception):
    """
    Base class for all exceptions that can be raised when using the openobd library.
    """

    def __init__(self, details="", status=-1, status_description=""):
        self.details = details
        self.status = status
        self.status_description = status_description

    def __str__(self):
        exception_info = self.__class__.__name__

        if self.details:
            exception_info += f": {self.details}"
        if self.status != -1:
            exception_info += f" (gRPC status: {self.status}"
            if self.status_description:
                exception_info += f", {self.status_description}"
            exception_info += ")"

        return exception_info

class OpenOBDStreamException(OpenOBDException):
    """
    Exception that can occur when handling gRPC streams.
    """
    pass


class OpenOBDStreamStoppedException(OpenOBDStreamException):
    pass


class OpenOBDStreamTimeoutException(OpenOBDStreamException):
    pass



def raises_openobd_exceptions(func):
    """
    If the wrapped function raises a gRPC exception, it will be cast and raised as an OpenOBDException.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            if isinstance(e, grpc.Call):
                # Encountered an exception raised by gRPC, so cast it to an OpenOBDException
                raise OpenOBDException(details=e.details(), status=e.code().value[0], status_description=e.code().value[1]) from None
            else:
                # The exception wasn't raised by gRPC, so just raise it as is
                raise e

    return wrapper



import functools
import threading
import typing
from queue import Queue, Empty
from collections.abc import Callable, Iterator, Iterable
from typing import Any
from grpc import Call, StatusCode
from openobd_protocol.Messages import Empty_pb2 as grpcEmpty

from openobd.core.exceptions import OpenOBDException, OpenOBDStreamException, OpenOBDStreamStoppedException, OpenOBDStreamTimeoutException

class IteratorStopped(Exception):
    pass

def requires_active_iterator(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        iterator = args[0]
        assert isinstance(iterator, MessageIterator), "Decorator 'requires_active_iterator' used outside of MessageIterator!"
        if not iterator.is_active:
            raise IteratorStopped("Unable to fulfill request, as the iterator has been stopped.")
        return func(*args, **kwargs)
    return wrapper


class MessageIterator:

    def __init__(self, message: Any = None):
        self.is_active = True

        self._next_messages = Queue()
        self._lock = threading.RLock()
        self._next_iteration_condition = threading.Condition(self._lock)
        self._queue_empty_condition = threading.Condition(self._lock)
        self._stopped_iteration_condition = threading.Condition(self._lock)
        if message is not None:
            self.send_message(message)

    def __iter__(self):
        return self

    def __next__(self):
        with self._lock:
            if not self.is_active:
                self._stopped_iteration_condition.notify_all()
                raise StopIteration

            if self._next_messages.qsize() == 0:
                self._queue_empty_condition.notify_all()
                self._next_iteration_condition.wait()
                if not self.is_active:
                    self._stopped_iteration_condition.notify_all()
                    raise StopIteration

            value = self._next_messages.get()
        return value

    @requires_active_iterator
    def send_message(self, message: Any):
        with self._lock:
            self._next_messages.put(message)
            self._next_iteration_condition.notify_all()

    @requires_active_iterator
    def stop(self, send_remaining_messages=False):
        with self._lock:
            if send_remaining_messages:
                self.wait_until_messages_sent()
            self.is_active = False
            self._next_iteration_condition.notify_all()
            # Wait until the iterator has raised a StopIteration exception before returning
            self._stopped_iteration_condition.wait(1)

    @requires_active_iterator
    def wait_until_messages_sent(self, timeout=10):
        with self._lock:
            if self._next_messages.qsize() == 0:
                return
            self._queue_empty_condition.wait(timeout)


class ThreadSafeVariable:

    def __init__(self, value=None):
        self._value = value
        self._lock = threading.Lock()

    @property
    def value(self):
        with self._lock:
            return self._value

    @value.setter
    def value(self, value):
        with self._lock:
            self._value = value


class StreamState:
    UNDEFINED = 0
    ACTIVE = 1
    SHUTTING_DOWN = 2   # The stream has received a cancel signal and cannot send messages anymore, but might still receive any final incoming messages
    CLOSED = 3


def requires_active_stream(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        stream_handler = args[0]
        assert isinstance(stream_handler, StreamHandler), "Decorator 'requires_active_stream' used outside of a StreamHandler class!"
        if stream_handler.received_exception:
            raise stream_handler.received_exception
        if stream_handler.stream_state != StreamState.ACTIVE:
            raise OpenOBDStreamStoppedException("Unable to fulfill request, as there is no stream active.")
        return func(*args, **kwargs)
    return wrapper


class StreamHandler:

    def __init__(self, stream_function: Callable[[Any], Any], request: Any = None, **kwargs):
        """
        Allows easier handling of a gRPC stream by providing methods to send and receive messages.

        :param stream_function: a reference to the function which should be used to send and receive messages.
        :param request: the message to send when not using an outgoing stream. Leave it None to send an EmptyMessage.

        :keyword outgoing_stream: boolean specifying whether the function requires a stream of messages as input. If not given, tries to detect it based on the function's type hints.
        """
        # Explicitly set whether an outgoing stream is used (if given as kwarg), or detect it based on type hints
        outgoing_stream = kwargs["outgoing_stream"] if "outgoing_stream" in kwargs else self._has_outgoing_stream(stream_function)
        assert isinstance(outgoing_stream, bool), "kwarg outgoing_stream must be a boolean."

        # self._request_iterator is only required for outgoing streams
        self._request_iterator = MessageIterator() if outgoing_stream else None

        # Determine the request that should be passed when starting the stream
        if outgoing_stream:
            request = self._request_iterator
        else:
            request = request if request is not None else grpcEmpty.EmptyMessage()

        # Set up required attributes (use ThreadSafeVariables for some to help prevent race conditions)
        self.__response_iterator = ThreadSafeVariable(None)     # Represents the incoming stream
        self._incoming_messages = Queue()    # Stores all incoming messages to later be retrieved
        self._stream_state = ThreadSafeVariable(StreamState.ACTIVE)
        self.received_exception = None

        # Start the thread containing the stream
        stream_thread = threading.Thread(target=self._stream, args=[stream_function, request], daemon=True)
        stream_thread.start()

    @property
    def _response_iterator(self):
        return self.__response_iterator.value

    @_response_iterator.setter
    def _response_iterator(self, value):
        self.__response_iterator.value = value

    @property
    def stream_state(self):
        return self._stream_state.value

    @stream_state.setter
    def stream_state(self, value):
        self._stream_state.value = value

    @staticmethod
    def _has_outgoing_stream(function: Callable[[Any], Any]) -> bool:
        type_hints = typing.get_type_hints(function)
        parameters = {k: v for k, v in type_hints.items() if k != "return"}
        parameter_type = next(iter(parameters.values()), None)
        # Check if any type hints were found before returning
        if parameter_type is None:
            raise OpenOBDStreamException("Unable to detect whether the given stream function requires an outgoing stream. Please pass it explicitly using the outgoing_stream kwarg.")
        return typing.get_origin(parameter_type) == Iterator

    def _get_incoming_message(self, block: bool = True, timeout: float | None = None) -> Any:
        try:
            incoming_message = self._incoming_messages.get(block=block, timeout=timeout)
            # If an exception occurred while waiting, raise it
            if isinstance(incoming_message, Exception):
                raise incoming_message
            else:
                return incoming_message
        except Empty:
            if block:
                raise OpenOBDStreamTimeoutException from None
            else:
                return None

    def receive(self, block: bool = True, timeout: float | None = None) -> Any:
        """
        Retrieves the oldest pending message. If there are no pending messages, wait until a new message arrives.
        Raises an OpenOBDStreamStoppedException if the stream is closed and there are no more pending messages.

        :param block: whether to wait for a new message if there are no pending messages. If False, returns None when no messages are pending.
        :param timeout: time in seconds to wait for a new message before raising an OpenOBDStreamTimeoutException. None will wait forever.
        :return: a message received by the gRPC stream.
        """
        # If an exception occurred in this stream, raise it
        if self.received_exception:
            raise self.received_exception

        # Check if there are currently any pending messages to receive (without spending any time waiting/blocking)
        incoming_message = self._get_incoming_message(block=False)
        if incoming_message is not None:
            return incoming_message

        if self.stream_state == StreamState.CLOSED:
            # All messages have been received and the stream is closed
            raise OpenOBDStreamStoppedException("The stream has been closed, so no further messages can be received.")

        # There are no pending messages, and the stream is still active, so wait for an incoming message as specified by the arguments
        return self._get_incoming_message(block=block, timeout=timeout)

    def flush_incoming_messages(self):
        """
        Discards all currently pending incoming messages.
        """
        with self._incoming_messages.mutex:
            self._incoming_messages.queue.clear()

    @requires_active_stream
    def send(self, message: Any, flush_incoming_messages: bool = False) -> None:
        """
        Sends a new message on this object's stream.

        :param message: the message to send on the stream.
        :param flush_incoming_messages: discards all currently pending incoming messages before sending the message.
        """
        if self._request_iterator is None:
            raise OpenOBDStreamException("It is not possible to send messages, as this object does not have an outgoing stream.")

        if flush_incoming_messages:
            self.flush_incoming_messages()

        self._request_iterator.send_message(message)

    @requires_active_stream
    def send_and_close(self, messages: Iterable, block: bool = True, timeout: float | None = None) -> Any:
        """
        Sends all the given messages on the stream, stops the stream after sending them, and returns the received
        response. This method is meant for calls that have an outgoing stream, and receive a single, unary response.

        :param messages: an iterable containing each message that should be sent on the stream.
        :param block: whether to wait for the response if there are no pending messages. If False, returns None when no messages are pending.
        :param timeout: time in seconds to wait for the response before raising an OpenOBDStreamTimeoutException. None will wait forever.
        :return: the response received from the gRPC stream.
        """
        if self._request_iterator is None:
            raise OpenOBDStreamException("It is not possible to send messages, as this object does not have an outgoing stream.")

        for message in messages:
            self.send(message)
        self.stop_stream(send_remaining_messages=True)
        return self.receive(block=block, timeout=timeout)

    def stop_stream(self, send_remaining_messages: bool = True) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new StreamHandler object will have to be created to start
        another stream.
        
        :param send_remaining_messages: whether to wait for all outgoing messages to be sent before stopping the stream.
        """
        # Indicate the stream is shutting down to prevent attempts at sending messages immediately after this call
        self.stream_state = StreamState.SHUTTING_DOWN

        if self._request_iterator is not None:
            # Stops the outgoing stream
            if self._request_iterator.is_active:
                self._request_iterator.stop(send_remaining_messages=send_remaining_messages)

        if self._response_iterator is not None:
            # Stops the incoming stream. If cancel is not present, it's fine
            if hasattr(self._response_iterator, "cancel"):   # Method defined by gRPC. Should always be present (sanity check).
                self._response_iterator.cancel()

    def _stream(self, stream_function: Callable[[Any], Any], request: Any) -> None:
        """
        Places any incoming messages in self._incoming_messages as they arrive.
        """
        try:
            # Start the stream by passing the request to the stream function
            response = stream_function(request)
            if isinstance(response, Iterator):
                # The server returned an iterator, which means that there is an incoming stream
                self._response_iterator = response
                # Check if the stream received a stop signal before it even started
                if self.stream_state != StreamState.ACTIVE:
                    self.stop_stream()
                # Handle the incoming stream
                for message in self._response_iterator:
                    self._incoming_messages.put(message)
            else:
                # There's no incoming stream, so merely pass on the response
                self._incoming_messages.put(response)
        except Exception as e:
            if (isinstance(e, Call) and e.code() == StatusCode.CANCELLED) or \
                    (isinstance(e, OpenOBDException) and e.status == 1):
                # The stream has been cancelled, so there's no need to stop it again or raise an exception
                pass
            else:
                self.stop_stream()

                if isinstance(e, Call):
                    # Encountered an exception raised by gRPC, so cast it to an OpenOBDStreamException
                    grpc_exception = OpenOBDStreamException(details=e.details(), status=e.code().value[0], status_description=e.code().value[1])
                    self.received_exception = grpc_exception
                else:
                    self.received_exception = e

                # In case receive() is currently waiting for an incoming message, pass them the exception
                self._incoming_messages.put(self.received_exception)
        finally:
            self.stream_state = StreamState.CLOSED

            # Put a StreamStoppedException in the queue to indicate to any blocking receive() calls that the stream has ended
            self._incoming_messages.put(OpenOBDStreamStoppedException("The stream has been closed, so no further messages can be received."))



import grpc

from openobd_protocol.Communication import CommunicationServices_pb2_grpc as grpcCommunicationService
from openobd_protocol.Configuration import ConfigurationServices_pb2_grpc as grpcConfigurationService
from openobd_protocol.ConnectionMonitor import ConnectionMonitorServices_pb2_grpc as grpcConnectionMonitorService
from openobd_protocol.Function import FunctionServices_pb2_grpc as grpcFunctionService
from openobd_protocol.Session import SessionServices_pb2_grpc as grpcService
from openobd_protocol.UserInterface import UserInterfaceServices_pb2_grpc as grpcUserInterfaceService
from openobd_protocol.SessionController import SessionControllerServices_pb2_grpc as grpcSessionControllerService
from openobd_protocol.FunctionBroker import FunctionBrokerServices_pb2_grpc as grpcFunctionBrokerServices
from openobd_protocol.VehicleInfo import VehicleInfoServices_pb2_grpc as grpcVehicleInfoServices

from .exceptions import OpenOBDException

class GrpcChannel:

    m_grpc_host = None
    m_grpc_port = 443
    m_connected = False
    m_channel = None

    def __init__(self, grpc_host, grpc_port):
        self.m_grpc_host = grpc_host
        self.m_grpc_port = grpc_port

    def connect(self):
        if not self.m_connected:
            ''' Check if local grpc-proxy is running '''
            if self.m_grpc_port == 443:
                self.m_channel = grpc.secure_channel(self.m_grpc_host, grpc.ssl_channel_credentials())
            else:
                ''' NOTE: Only use this for development purposes '''
                self.m_channel = grpc.insecure_channel('{}:{}'.format(self.m_grpc_host, self.m_grpc_port))

            ''' TODO: Exceptions and stuff '''
            self.m_connected = True

    def getChannel(self):
        if not self.m_connected:
            raise OpenOBDException("Grpc channel not connected")

        return self.m_channel

class GrpcFactory:
    def get_session(self) -> grpcService.sessionStub | grpcService.sessionServicer:
        return grpcService.sessionServicer()

    def get_function(self) -> grpcFunctionService.functionStub | grpcFunctionService.functionServicer:
        return grpcFunctionService.functionServicer()

    def get_config(self) -> grpcConfigurationService.configStub | grpcConfigurationService.configServicer:
        return grpcConfigurationService.configServicer()

    def get_can(self) -> grpcCommunicationService.canStub | grpcCommunicationService.canServicer:
        return grpcCommunicationService.canServicer()

    def get_kline(self) -> grpcCommunicationService.klineStub | grpcCommunicationService.klineServicer:
        return grpcCommunicationService.klineServicer()

    def get_terminal15(self) -> grpcCommunicationService.terminal15Stub | grpcCommunicationService.terminal15Servicer:
        return grpcCommunicationService.terminal15Servicer()

    def get_doip(self) -> grpcCommunicationService.doipStub | grpcCommunicationService.doipServicer:
        return grpcCommunicationService.doipServicer()

    def get_ui(self) -> grpcUserInterfaceService.userInterfaceStub | grpcUserInterfaceService.userInterfaceServicer:
        return grpcUserInterfaceService.userInterfaceServicer()

    def get_connection_monitor(self) -> grpcConnectionMonitorService.connectionMonitorStub | grpcConnectionMonitorService.connectionMonitorServicer:
        return grpcConnectionMonitorService.connectionMonitorServicer()

    def get_session_controller(self) -> grpcSessionControllerService.sessionControllerStub | grpcSessionControllerService.sessionControllerServicer:
        return grpcSessionControllerService.sessionControllerServicer()

    def get_function_broker(self) -> grpcFunctionBrokerServices.functionBrokerStub | grpcFunctionBrokerServices.functionBrokerServicer:
        return grpcFunctionBrokerServices.functionBrokerServicer()

    def get_vehicle_info(self) -> grpcVehicleInfoServices.vehicleInfoStub | grpcVehicleInfoServices.vehicleInfoServicer:
        return grpcVehicleInfoServices.vehicleInfoServicer()

class NetworkGrpcFactory(GrpcFactory):

    def __init__(self, grpc_host=None, grpc_port=443):
        """
        A factory that returns the methods to make grpc calls via the network.
        :param grpc_host: the hostname for the grpc server
        :param grpc_port: the default value of 443 should be used to make gRPC calls using SSL.
        """
        self.grpc_channel = GrpcChannel(grpc_host, grpc_port)
        self.grpc_channel.connect()

    def get_session(self) -> grpcService.sessionStub | grpcService.sessionServicer:
        return grpcService.sessionStub(self.grpc_channel.getChannel())

    def get_function(self) -> grpcFunctionService.functionStub | grpcFunctionService.functionServicer:
        return grpcFunctionService.functionStub(self.grpc_channel.getChannel())

    def get_config(self) -> grpcConfigurationService.configStub | grpcConfigurationService.configServicer:
        return grpcConfigurationService.configStub(self.grpc_channel.getChannel())

    def get_can(self) -> grpcCommunicationService.canStub | grpcCommunicationService.canServicer:
        return grpcCommunicationService.canStub(self.grpc_channel.getChannel())

    def get_kline(self) -> grpcCommunicationService.klineStub | grpcCommunicationService.klineServicer:
        return grpcCommunicationService.klineStub(self.grpc_channel.getChannel())

    def get_terminal15(self) -> grpcCommunicationService.terminal15Stub | grpcCommunicationService.terminal15Servicer:
        return grpcCommunicationService.terminal15Stub(self.grpc_channel.getChannel())

    def get_doip(self) -> grpcCommunicationService.doipStub | grpcCommunicationService.doipServicer:
        return grpcCommunicationService.doipStub(self.grpc_channel.getChannel())

    def get_ui(self) -> grpcUserInterfaceService.userInterfaceStub | grpcUserInterfaceService.userInterfaceServicer:
        return grpcUserInterfaceService.userInterfaceStub(self.grpc_channel.getChannel())

    def get_connection_monitor(self) -> grpcConnectionMonitorService.connectionMonitorStub | grpcConnectionMonitorService.connectionMonitorServicer:
        return grpcConnectionMonitorService.connectionMonitorStub(self.grpc_channel.getChannel())

    def get_session_controller(self) -> grpcSessionControllerService.sessionControllerStub | grpcSessionControllerService.sessionControllerServicer:
        return grpcSessionControllerService.sessionControllerStub(self.grpc_channel.getChannel())

    def get_function_broker(self) -> grpcFunctionBrokerServices.functionBrokerStub | grpcFunctionBrokerServices.functionBrokerServicer:
        return grpcFunctionBrokerServices.functionBrokerStub(self.grpc_channel.getChannel())

    def get_vehicle_info(self) -> grpcVehicleInfoServices.vehicleInfoStub | grpcVehicleInfoServices.vehicleInfoServicer:
        return grpcVehicleInfoServices.vehicleInfoStub(self.grpc_channel.getChannel())


from .connection_monitor import *
from .response_exceptions import *
from .isotp_socket import *
from .kline_socket import *
from .raw_socket import *
from .terminal15_socket import *
from .doip_socket import *



from openobd.core.exceptions import OpenOBDStreamException


class ResponseException(OpenOBDStreamException):
    """
    Base class for all exceptions that can be raised during vehicle communication.
    """

    def __init__(self, request="", response="", request_id=0, response_id=0, **kwargs):
        self.request = request
        self.response = response
        self.request_id = request_id
        self.response_id = response_id
        super().__init__(**kwargs)

    def __str__(self):
        exception_info = self.__class__.__name__

        if self.request_id:
            exception_info += f" ({self.request_id:X}"
            if self.response_id:
                exception_info += f" -> {self.response_id:X}"
            exception_info += ")"
        if self.request:
            exception_info += f" request: {self.request}"
        if self.response:
            exception_info += f" response: {self.response}"

        return exception_info


class NoResponseException(ResponseException):
    """
    Did not receive a response from the vehicle in the specified time.
    """
    pass


class InvalidResponseException(ResponseException):
    """
    The response received from the vehicle does not have the correct format.
    """
    pass


class NegativeResponseException(ResponseException):
    """
    Base class for all exceptions that are raised because of a negative response received from the vehicle.
    """
    pass


class GeneralRejectException(NegativeResponseException):
    pass


class ServiceNotSupportedException(NegativeResponseException):
    pass


class SubFunctionNotSupportedException(NegativeResponseException):
    """
    For KWP2000, this exception means "subFunctionNotSupported-invalidFormat" according to ISO 14230-3:1996.
    """
    pass


class IncorrectMessageLengthOrInvalidFormatException(NegativeResponseException):
    pass


class ResponseTooLongException(NegativeResponseException):
    pass


class BusyRepeatRequestException(NegativeResponseException):
    pass


class ConditionsNotCorrectException(NegativeResponseException):
    """
    For KWP2000, this exception means "conditionsNotCorrect or requestSequenceError" according to ISO 14230-3:1996.
    """
    pass


class RequestSequenceErrorException(NegativeResponseException):
    pass


class NoResponseFromSubnetComponentException(NegativeResponseException):
    pass


class FailurePreventsExecutionOfRequestedActionException(NegativeResponseException):
    pass


class RequestOutOfRangeException(NegativeResponseException):
    pass


class SecurityAccessDeniedException(NegativeResponseException):
    pass


class AuthenticationRequiredException(NegativeResponseException):
    pass


class InvalidKeyException(NegativeResponseException):
    pass


class ExceedNumberOfAttemptsException(NegativeResponseException):
    pass


class RequiredTimeDelayNotExpiredException(NegativeResponseException):
    pass


class SecureDataTransmissionRequiredException(NegativeResponseException):
    pass


class SecureDataTransmissionNotAllowedException(NegativeResponseException):
    pass


class SecureDataVerificationFailedException(NegativeResponseException):
    pass


class CertificateVerificationFailedException(NegativeResponseException):
    """
    Base class for certificate verification negative response codes.
    """
    pass


class CertificateVerificationFailedInvalidTimePeriodException(CertificateVerificationFailedException):
    """
    For KWP2000, this exception means "uploadNotAccepted" according to ISO 14230-3:1996.
    """
    pass


class CertificateVerificationFailedInvalidSignatureException(CertificateVerificationFailedException):
    """
    For KWP2000, this exception means "improperUploadType" according to ISO 14230-3:1996.
    """
    pass


class CertificateVerificationFailedInvalidChainOfTrustException(CertificateVerificationFailedException):
    """
    For KWP2000, this exception means "can'tUploadFromSpecifiedAddress" according to ISO 14230-3:1996.
    """
    pass


class CertificateVerificationFailedInvalidTypeException(CertificateVerificationFailedException):
    """
    For KWP2000, this exception means "can'tUploadNumberOfBytesRequested" according to ISO 14230-3:1996.
    """
    pass


class CertificateVerificationFailedInvalidFormatException(CertificateVerificationFailedException):
    pass


class CertificateVerificationFailedInvalidContentException(CertificateVerificationFailedException):
    pass


class CertificateVerificationFailedInvalidScopeException(CertificateVerificationFailedException):
    pass


class CertificateVerificationFailedInvalidCertificateRevokedException(CertificateVerificationFailedException):
    pass


class OwnershipVerificationFailedException(CertificateVerificationFailedException):
    pass


class ChallengeCalculationFailedException(CertificateVerificationFailedException):
    pass


class SettingAccessRightsFailedException(CertificateVerificationFailedException):
    pass


class SessionKeyCreationDerivationFailedException(CertificateVerificationFailedException):
    pass


class ConfigurationDataUsageFailedException(CertificateVerificationFailedException):
    pass


class DeAuthenticationFailedException(CertificateVerificationFailedException):
    pass


class UploadDownloadNotAcceptedException(NegativeResponseException):
    pass


class TransferDataSuspendedException(NegativeResponseException):
    """
    For KWP2000, this exception means "transferSuspended" according to ISO 14230-3:1996.
    """
    pass


class GeneralProgrammingFailureException(NegativeResponseException):
    """
    For KWP2000, this exception means "transferAborted" according to ISO 14230-3:1996.
    """
    pass


class WrongBlockSequenceCounterException(NegativeResponseException):
    pass


class RequestCorrectlyReceivedResponsePendingException(NegativeResponseException):
    pass


class SubFunctionNotSupportedInActiveSessionException(NegativeResponseException):
    pass


class ServiceNotSupportedInActiveSessionException(NegativeResponseException):
    pass


class SpecificConditionNotCorrectException(NegativeResponseException):
    """
    Base class for specific conditions negative response codes.
    """
    pass


class RpmTooHighException(SpecificConditionNotCorrectException):
    pass


class RpmTooLowException(SpecificConditionNotCorrectException):
    pass


class EngineIsRunningException(SpecificConditionNotCorrectException):
    pass


class EngineIsNotRunningException(SpecificConditionNotCorrectException):
    pass


class EngineRunTimeTooLowException(SpecificConditionNotCorrectException):
    pass


class TemperatureTooHighException(SpecificConditionNotCorrectException):
    pass


class TemperatureTooLowException(SpecificConditionNotCorrectException):
    pass


class VehicleSpeedTooHighException(SpecificConditionNotCorrectException):
    pass


class VehicleSpeedTooLowException(SpecificConditionNotCorrectException):
    pass


class ThrottlePedalTooHighException(SpecificConditionNotCorrectException):
    pass


class ThrottlePedalTooLowException(SpecificConditionNotCorrectException):
    pass


class TransmissionRangeNotInNeutralException(SpecificConditionNotCorrectException):
    pass


class TransmissionRangeNotInGearException(SpecificConditionNotCorrectException):
    pass


class BrakeSwitchesNotClosedException(SpecificConditionNotCorrectException):
    pass


class ShifterLeverNotInParkException(SpecificConditionNotCorrectException):
    pass


class TorqueConverterClutchLockedException(SpecificConditionNotCorrectException):
    pass


class VoltageTooHighException(SpecificConditionNotCorrectException):
    pass


class VoltageTooLowException(SpecificConditionNotCorrectException):
    pass


class ResourceTemporarilyNotAvailableException(SpecificConditionNotCorrectException):
    pass


class UnknownNegativeResponseException(NegativeResponseException):
    pass


# Based on ISO 14229-1:2020
negative_response_code_exceptions = {
    # 01-0F: ISOSAEReserved
    "10": GeneralRejectException,
    "11": ServiceNotSupportedException,
    "12": SubFunctionNotSupportedException,
    "13": IncorrectMessageLengthOrInvalidFormatException,
    "14": ResponseTooLongException,
    # 15-20: ISOSAEReserved
    "21": BusyRepeatRequestException,
    "22": ConditionsNotCorrectException,
    # 23: ISOSAEReserved
    "24": RequestSequenceErrorException,
    "25": NoResponseFromSubnetComponentException,
    "26": FailurePreventsExecutionOfRequestedActionException,
    # 27-30: ISOSAEReserved
    "31": RequestOutOfRangeException,
    # 32: ISOSAEReserved
    "33": SecurityAccessDeniedException,
    "34": AuthenticationRequiredException,
    "35": InvalidKeyException,
    "36": ExceedNumberOfAttemptsException,
    "37": RequiredTimeDelayNotExpiredException,
    "38": SecureDataTransmissionRequiredException,
    "39": SecureDataTransmissionNotAllowedException,
    "3A": SecureDataVerificationFailedException,
    # 3B-4F: ISOSAEReserved
    "50": CertificateVerificationFailedInvalidTimePeriodException,
    "51": CertificateVerificationFailedInvalidSignatureException,
    "52": CertificateVerificationFailedInvalidChainOfTrustException,
    "53": CertificateVerificationFailedInvalidTypeException,
    "54": CertificateVerificationFailedInvalidFormatException,
    "55": CertificateVerificationFailedInvalidContentException,
    "56": CertificateVerificationFailedInvalidScopeException,
    "57": CertificateVerificationFailedInvalidCertificateRevokedException,
    "58": OwnershipVerificationFailedException,
    "59": ChallengeCalculationFailedException,
    "5A": SettingAccessRightsFailedException,
    "5B": SessionKeyCreationDerivationFailedException,
    "5C": ConfigurationDataUsageFailedException,
    "5D": DeAuthenticationFailedException,
    # 5E-6F: ISOSAEReserved
    "70": UploadDownloadNotAcceptedException,
    "71": TransferDataSuspendedException,
    "72": GeneralProgrammingFailureException,
    "73": WrongBlockSequenceCounterException,
    # 74-77: ISOSAEReserved
    "78": RequestCorrectlyReceivedResponsePendingException,
    # 79-7D: ISOSAEReserved
    "7E": SubFunctionNotSupportedInActiveSessionException,
    "7F": ServiceNotSupportedInActiveSessionException,
    # 80: ISOSAEReserved
    "81": RpmTooHighException,
    "82": RpmTooLowException,
    "83": EngineIsRunningException,
    "84": EngineIsNotRunningException,
    "85": EngineRunTimeTooLowException,
    "86": TemperatureTooHighException,
    "87": TemperatureTooLowException,
    "88": VehicleSpeedTooHighException,
    "89": VehicleSpeedTooLowException,
    "8A": ThrottlePedalTooHighException,
    "8B": ThrottlePedalTooLowException,
    "8C": TransmissionRangeNotInNeutralException,
    "8D": TransmissionRangeNotInGearException,
    # 8E: ISOSAEReserved
    "8F": BrakeSwitchesNotClosedException,
    "90": ShifterLeverNotInParkException,
    "91": TorqueConverterClutchLockedException,
    "92": VoltageTooHighException,
    "93": VoltageTooLowException,
    "94": ResourceTemporarilyNotAvailableException,
    # 95-EF: reservedForSpecificConditionsNotCorrect
    # F0-FE: vehicleManufacturerSpecificConditionsNotCorrect
    # FF: ISOSAEReserved
}



from openobd_protocol.Function.Messages.Function_pb2 import FunctionContext
from openobd_protocol.Session.Messages.ServiceResult_pb2 import Result, UserFeedback

from openobd.core.exceptions import OpenOBDException


class OpenOBDFunctionException(OpenOBDException):
    """
    Base class for all exceptions that can be raised when running functions.
    """

    def __init__(self, function_id: str = None, function_context: FunctionContext = None, **kwargs):
        self.function_id = function_id
        self.function_context = function_context
        super().__init__(**kwargs)

    def __str__(self):
        exception_data = []

        if self.function_id:
            exception_data.append(f"Function ID [{self.function_id}].")

        if self.function_context:
            if self.function_context.service_result.result:
                results = []
                for result in self.function_context.service_result.result:
                    results.append(Result.Name(result))
                results_string = ", ".join(results)
                exception_data.append(f"Result(s): {results_string}.")
            else:
                exception_data.append("No result set.")

            if self.function_context.service_result.user_feedback:
                exception_data.append(f"User feedback: {UserFeedback.Name(self.function_context.service_result.user_feedback)}.")

        exception_string = self.__class__.__name__
        if len(exception_data) != 0:
            exception_string += ": " + " ".join(exception_data)

        return exception_string


class OpenOBDFunctionUnsuccessfulException(OpenOBDFunctionException):
    """
    The result of the function is not only set to success.
    """
    pass


class OpenOBDFunctionDeadlineExceededException(OpenOBDFunctionException):
    """
    The requested function did not respond in time.
    """
    pass


class OpenOBDFunctionPermissionDeniedException(OpenOBDFunctionException):
    """
    You are not allowed to execute the requested function.
    """
    pass


class OpenOBDFunctionUnavailableException(OpenOBDFunctionException):
    """
    The requested function is not available.
    """
    pass



from openobd_protocol.Messages import Empty_pb2 as grpcEmpty

import threading

from openobd.core.session import OpenOBDSession
from openobd.core.exceptions import OpenOBDException
from openobd.core.stream_handler import StreamHandler
from openobd.core.exceptions import OpenOBDStreamStoppedException

class SessionTokenHandler:

    def __init__(self, openobd_session: OpenOBDSession):
        """
        Authenticates the given openOBD session and starts a thread to keep the session active by periodically updating
        its session token.

        :param openobd_session: a newly created OpenOBDSession that should remain active.
        """
        self.openobd_session = openobd_session
        self._authenticate()
        self.refresh_session_token_thread = threading.Thread(target=self._refresh_session_token, daemon=True)
        self.refresh_session_token_thread.start()

    def _authenticate(self):
        # Authenticating the openOBD session
        new_token = self.openobd_session.authenticate(request=grpcEmpty.EmptyMessage())
        self.openobd_session.update_session_token(new_token.value)

        # Start refreshing session tokens
        self.session_token_manager = StreamHandler(self.openobd_session.open_session_token_stream)

    def _refresh_session_token(self):
        try:
            while True:
                new_token = self.session_token_manager.receive()
                self.openobd_session.update_session_token(new_token.value)
        except OpenOBDStreamStoppedException:
            pass
        except OpenOBDException as e:
            print(f" [!] Stopped refreshing session tokens for session [{self.openobd_session.id()}] due to an exception.")
            print(e)
        finally:
            self.session_token_manager.stop_stream()



from openobd_protocol.Communication.Messages.Raw_pb2 import RawChannel, RawFrame

from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler


class RawSocket:

    def __init__(self, openobd_session: OpenOBDSession, raw_channel: RawChannel, timeout: float | None = 10):
        """
        Handles sending and receiving frame payloads to and from the given channel.

        :param openobd_session: an active OpenOBDSession with which to start the raw frames stream.
        :param raw_channel: the channel which to exchange frames with.
        :param timeout: time in seconds to wait for an incoming frame before raising a OpenOBDStreamTimeoutException. None will wait forever.
        """
        self.stream_handler = StreamHandler(openobd_session.open_raw_stream, outgoing_stream=True)
        self._channel = raw_channel
        self.timeout = timeout
        '''We send an empty message to the server to ensure that the channel is initialized, this also helps when we 
        want to just listen on a channel'''
        self.stream_handler.send(RawFrame(channel=self._channel, payload=""))

    def send(self, payload: str, flush_incoming_messages: bool = False) -> None:
        """
        Sends the given payload to the channel.

        :param payload: the payload of the frame to send to the channel.
        :param flush_incoming_messages: discards all currently pending incoming frames before sending the payload.
        """
        message = RawFrame(channel=self._channel, payload=payload)
        self.stream_handler.send(message, flush_incoming_messages)

    def receive(self, block: bool = True, timeout: float | None = None) -> str | None:
        """
        Retrieves the oldest pending incoming frame. If there are no pending frames, wait until a new frame arrives.

        :param block: whether to wait for a new frame if there are no pending frames. If False, returns None when no frames are pending.
        :param timeout: time in seconds to wait for a new frame before raising a OpenOBDStreamTimeoutException. Will use self.timeout if None.
        :return: the payload received from the channel.
        """
        timeout = timeout if timeout is not None else self.timeout
        response = self.stream_handler.receive(block, timeout)
        if response is not None:
            return response.payload.upper()

    def stop_stream(self) -> None:
        """
        Closes the gRPC stream if it is not already closed. A new RawSocket object will have to be created to start
        another stream.
        """
        self.stream_handler.stop_stream()
